﻿using System;
using System.Runtime.InteropServices;
using System.Drawing;

namespace ACLAS
{
    public enum CDisplayModel
    {
        VFD1x9,
        VFD24x128,
        LCD32x240,
        LCD32x144,
        LCD65x132,
        LCD65x264,
        COG64x384
    }

    public class CDisplayClass
    {
        private Device SubLCD = null;
        private bool _GraphicsMode = false;
        private int _Width, _Height;
        private int bSize;

        private const String DllName = "tvout.dll";


        [DllImport(DllName)]
        private static extern Boolean CLCDImage(IntPtr hbm, int Width, int Height, Byte[] buf);

        private Byte[] CLCDBuffer;

        Bitmap bmp = null;

        public Graphics gr = null;

        public Brush brush = new SolidBrush(Color.Black);

        public CDisplayModel Model = CDisplayModel.LCD32x144;

        public bool GraphicsMode
        {
            get
            {
                return _GraphicsMode;
            }
        }

        public int Width
        {
            get
            {
                return _Width;
            }
        }

        public int Height
        {
            get
            {
                return _Height;
            }
        }

        public CDisplayClass()
        {
            // 打開驅動裝置類別
            SubLCD = new Device("SLD1:", WINAPI.GENERIC_WRITE);
        }

        public bool Open()
        {
            return SubLCD.Open();
        }

        public void Close()
        {
            SubLCD.Close();
            if (gr != null)
                gr.Dispose();
            if (bmp != null)
                bmp.Dispose();
        }

        public void GetLCDInfo(ref int col, ref int row)
        {
            byte[] buf = new byte[2];
            int rsize = 0;
            SubLCD.IOControl((int)SLD_IOCTRL.GETDOTCOL, null, 0, buf, 2, ref rsize);
            col = buf[0] | (buf[1] << 8);

            SubLCD.IOControl((int)SLD_IOCTRL.GETDOTROW, null, 0, buf, 2, ref rsize);
            row = buf[0] | (buf[1] << 8);

            _Width = col;
            _Height = row;

            if (col == 240 && row == 32)
            {
                Model = CDisplayModel.LCD32x240;
                _GraphicsMode = true;
            }
            else if (col == 9 && row == 1)
            {
                Model = CDisplayModel.VFD1x9;
            }
            else if (col == 128 && row == 24)
            {
                Model = CDisplayModel.VFD24x128;
                _GraphicsMode = true;
            }
            else if (col == 132 && row == 65)
            {
                Model = CDisplayModel.LCD65x132;
                _GraphicsMode = true;
            }
            else if (col == 144 && row == 32)
            {
                Model = CDisplayModel.LCD32x144;
                _GraphicsMode = true;
            }
            else if (col == 264 && row == 65)
            {
                Model = CDisplayModel.LCD65x264;
                _GraphicsMode = true;
            }
            else if (col == 384 && row == 64)
            {
                Model = CDisplayModel.COG64x384;
                _GraphicsMode = true;
            }

            if (_GraphicsMode)
            {
                if (bmp != null)
                {
                    bmp.Dispose();
                    gr.Dispose();
                }
                bmp = new Bitmap(Width, Height, System.Drawing.Imaging.PixelFormat.Format16bppRgb565);
                gr = Graphics.FromImage(bmp);
                Clear();
                bSize = ((col + 7) / 8) * row + 16;
                CLCDBuffer = new Byte[bSize];
                Show();
            }

            if (Model == CDisplayModel.LCD32x144 || Model == CDisplayModel.LCD65x132)
            {
                SetGraphMode(false);
            }
        }

        public void Clear()
        {
            if (_GraphicsMode)
            {
                gr.Clear(Color.White);
            }
        }

        public void Show()
        {
            if (_GraphicsMode)
            {
                CLCDImage(bmp.GetHbitmap(), Width, Height, CLCDBuffer);
                SubLCD.Write(CLCDBuffer, bSize);
            }
        }

        public void BackLightOff()
        {
            byte[] buf = { 0 };
            int rsize = 0;
            SubLCD.IOControl((int)SLD_IOCTRL.LIGHT_OP, buf, 1, null, 0, ref rsize);
        }

        public void BackLightOn()
        {
            byte[] buf = { 1 };
            int rsize = 0;
            SubLCD.IOControl((int)SLD_IOCTRL.LIGHT_OP, buf, 1, null, 0, ref rsize);
        }

        public void SetLCDContrast(int vref, int ratio)
        {
            byte[] buf = new byte[2];
            buf[0] = (byte)vref;
            buf[1] = (byte)ratio;
            int rsize = 0;
            SubLCD.IOControl((int)SLD_IOCTRL.SET_CONTRAST, buf, 2, null, 0, ref rsize);
        }

        public void SetLCDContrast(int grad)
        {
            byte[] buf = new byte[1];
            buf[0] = (byte)grad;
            int rsize = 0;
            SubLCD.IOControl((int)SLD_IOCTRL.SET_CONTRAST, buf, 1, null, 0, ref rsize);
        }

        public bool GetLCDContrast(ref int grad)
        {
            byte[] buf = new byte[1];
            int rsize = 0;
            if (SubLCD.IOControl((int)SLD_IOCTRL.GET_CONTRAST, null, 0, buf, 1, ref rsize))
            {
                grad = buf[0];
                return true;
            }
            else
                return false;
        }

        public bool GetLCDContrastRange(ref int gradrange)
        {
            byte[] buf = new byte[1];
            int rsize = 0;
            if (SubLCD.IOControl((int)SLD_IOCTRL.GET_CONTRAST_RANGE, null, 0, buf, 1, ref rsize))
            {
                gradrange = buf[0];
                return true;
            }
            else
                return false;
        }

        public void ClrSubLCD()  // 重新啟始螢幕
        {
            if (!_GraphicsMode)
            {
                Byte[] clr = new Byte[] { 27, (Byte)'@' };
                SubLCD.Write(clr, 2);
            }
            else
            {
                Clear();
                Show();
            }
        }

        public void Show(Byte[] buf, int len)
        {
            SubLCD.Write(buf, len);
        }

        public void SetGraphMode(bool mode)
        {
            byte[] buf = new byte[4];
            int val = mode ? 1 : 0;
            int rsize = 0;
            CovTool.IntToByte(val, buf, 0);
            SubLCD.IOControl((int)SLD_IOCTRL.SET_GRAPHMODE, buf, 4, null, 0, ref rsize);
            _GraphicsMode = mode;
        }
    }
}