﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Collections;
using System.Threading;


namespace ACLAS
{
    public class ConnectManager
    {
        const int S_OK = 0;
        const uint CONNMGR_PARAM_GUIDDESTNET = 0x1;
        const uint CONNMGR_PRIORITY_USERINTERACTIVE = 0x08000;
        const uint INFINITE = 0xffffffff;
        const uint CONNMGR_STATUS_CONNECTED = 0x10;
        const int CONNMGR_MAX_DESC = 128;    // @constdefine Max size of a network description

        const int CONNMGR_FLAG_PROXY_HTTP = 0x1; // @constdefine HTTP Proxy supported
        const int CONNMGR_FLAG_PROXY_WAP = 0x2; // @constdefine WAP Proxy (gateway) supported
        const int CONNMGR_FLAG_PROXY_SOCKS4 = 0x4; // @constdefine SOCKS4 Proxy supported
        const int CONNMGR_FLAG_PROXY_SOCKS5 = 0x8; // @constdefine SOCKS5 Proxy supported

        const UInt16 IDC_WAIT = 32514;
        const UInt16 IDC_ARROW = 32512;

        private IntPtr m_hConnection = IntPtr.Zero;

        [DllImport("coredll.dll")]
        public static extern uint GetTickCount();

        [DllImport("coredll.dll")]
        public static extern uint WaitForSingleObject(IntPtr hHandle, uint dwMilliseconds);

        [DllImport("cellcore.dll")]
        public static extern int ConnMgrMapURL(string pwszURL, ref GUID pguid, ref uint pdwIndex);

        [DllImport("cellcore.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        unsafe static extern int ConnMgrMapURL(string url, ref Guid guid, int* nIndex);

        [DllImport("cellcore.dll")]
        public static extern int ConnMgrEstablishConnectionSync(ref CONNMGR_CONNECTIONINFO ci, ref IntPtr phConnection, uint dwTimeout, ref uint pdwStatus);

        [DllImport("cellcore.dll")]
        private static extern IntPtr ConnMgrApiReadyEvent();

        [DllImport("cellcore.dll")]
        public static extern int ConnMgrReleaseConnection(IntPtr hConnection, int bCache);

        [DllImport("cellcore.dll")]
        public static extern int ConnMgrEnumDestinations(int nIndex, ref CONNMGR_DESTINATION_INFO pDestInfo);

        [DllImport("cellcore.dll")]
        public static extern int ConnMgrConnectionStatus(
            IntPtr hConnection,    // @parm Handle to connection, returned from ConnMgrEstablishConnection
            ref uint pdwStatus       // @parm Returns current connection status, one of CONNMGR_STATUS_*
        );

        [DllImport("coredll.dll")]
        private static extern int CloseHandle(IntPtr hObject);

        public ConnectManager()
        {
        }

        ~ConnectManager()
        {
            ReleaseConnection();
        }

        /// <summary>
        /// 查看连接是否可用
        /// </summary>
        /// <returns></returns>
        public bool GetConnMgrAvailable()
        {
            IntPtr hConnMgr = ConnMgrApiReadyEvent();
            bool bAvailbale = false;
            uint dwResult = WaitForSingleObject(hConnMgr, 2000);

            if (dwResult == 0)
            {
                bAvailbale = true;
            }

            // 关闭
            if (hConnMgr.ToInt32() != 0) CloseHandle(hConnMgr);

            return bAvailbale;
        }
        /// <summary>
        /// 映射URL
        /// </summary>
        /// <param name="lpszURL"></param>
        /// <param name="guidNetworkObject"></param>
        /// <param name="pcsDesc"></param>
        /// <returns></returns>
        public int MapURLAndGUID(string lpszURL, ref GUID guidNetworkObject, ref string pcsDesc)
        {
            if (lpszURL == null || lpszURL.Length < 1)
                return 0;

            uint nIndex = 0;
            int hResult = ConnMgrMapURL(lpszURL, ref guidNetworkObject, ref nIndex);
            if (hResult < 0)
            {
                throw new Exception("Could not map a request to a network identifier");
            }
            else
            {
                if (pcsDesc != null)
                {
                    CONNMGR_DESTINATION_INFO DestInfo = new CONNMGR_DESTINATION_INFO();
                    if (ConnMgrEnumDestinations((int)nIndex, ref DestInfo) >= 0)
                    {
                        pcsDesc = DestInfo.szDescription;
                    }
                }
            }

            return (int)nIndex;
        }

        /// <summary>
        /// 枚举网络标识符信息
        /// </summary>
        /// <param name="lstNetIdentifiers"></param>
        public List<CONNMGR_DESTINATION_INFO> EnumNetIdentifier()
        {
            List<CONNMGR_DESTINATION_INFO> lstNetIdentifiers = new List<CONNMGR_DESTINATION_INFO>();
            // 得到网络列表
            for (uint dwEnumIndex = 0; ; dwEnumIndex++)
            {
                CONNMGR_DESTINATION_INFO networkDestInfo = new CONNMGR_DESTINATION_INFO();

                if (ConnMgrEnumDestinations((int)dwEnumIndex, ref networkDestInfo) != 0)
                {
                    break;
                }
                lstNetIdentifiers.Add(networkDestInfo);
            }

            return lstNetIdentifiers;
        }

        /// <summary>
        /// 建立连接
        /// </summary>
        /// <param name="nIndex"></param>
        /// <returns></returns>
        public bool EstablishConnection(uint nIndex, uint dwStatus)
        {
            ReleaseConnection();
            // 得到正确的连接信息
            CONNMGR_DESTINATION_INFO DestInfo = new CONNMGR_DESTINATION_INFO();
            int hResult = ConnMgrEnumDestinations((int)nIndex, ref DestInfo);

            if (hResult >= 0)
            {
                // 初始化连接结构
                CONNMGR_CONNECTIONINFO ConnInfo = new CONNMGR_CONNECTIONINFO();

                ConnInfo.cbSize = (uint)Marshal.SizeOf(ConnInfo);
                ConnInfo.dwParams = CONNMGR_PARAM_GUIDDESTNET;
                ConnInfo.dwFlags = CONNMGR_FLAG_PROXY_HTTP | CONNMGR_FLAG_PROXY_WAP | CONNMGR_FLAG_PROXY_SOCKS4 | CONNMGR_FLAG_PROXY_SOCKS5;
                ConnInfo.dwPriority = CONNMGR_PRIORITY_USERINTERACTIVE;
                ConnInfo.guidDestNet = DestInfo.guid;
                ConnInfo.bExclusive = 0;
                ConnInfo.bDisabled = 0;

                //uint dwStatus = 0;
                hResult = ConnMgrEstablishConnectionSync(ref ConnInfo, ref m_hConnection, 15 * 1000, ref dwStatus);
                if (hResult < 0)
                {
                    return false;
                }
                else
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// 建立连接
        /// </summary>
        /// <param name="nIndex"></param>
        /// <returns></returns>
        public bool EstablishConnection(GUID guid, uint dwStatus)
        {
            ReleaseConnection();
            // 得到正确的连接信息

            // 初始化连接结构
            CONNMGR_CONNECTIONINFO ConnInfo = new CONNMGR_CONNECTIONINFO();

            ConnInfo.cbSize = (uint)Marshal.SizeOf(ConnInfo);
            ConnInfo.dwParams = CONNMGR_PARAM_GUIDDESTNET;
            ConnInfo.dwFlags = CONNMGR_FLAG_PROXY_HTTP | CONNMGR_FLAG_PROXY_WAP | CONNMGR_FLAG_PROXY_SOCKS4 | CONNMGR_FLAG_PROXY_SOCKS5;
            ConnInfo.dwPriority = CONNMGR_PRIORITY_USERINTERACTIVE;
            ConnInfo.guidDestNet = guid;
            ConnInfo.bExclusive = 0;
            ConnInfo.bDisabled = 0;

            //uint dwStatus = 0;
            int hResult = ConnMgrEstablishConnectionSync(ref ConnInfo, ref m_hConnection, 15 * 1000, ref dwStatus);
            if (hResult < 0)
            {
                return false;
            }
            else
            {
                return true;
            }
        }

        /// <summary>
        /// 连接状态
        /// </summary>
        /// <param name="nTimeoutSec"></param>
        /// <param name="pdwStatus"></param>
        /// <returns></returns>
        public bool WaitForConnected(int nTimeoutSec, ref uint pdwStatus)
        {
            uint dwStartTime = GetTickCount();
            bool bRet = false;

            while (GetTickCount() - dwStartTime < (uint)nTimeoutSec * 1000)
            {
                if (m_hConnection.ToInt32() != 0)
                {
                    uint dwStatus = 0;
                    int hr = ConnMgrConnectionStatus(m_hConnection, ref dwStatus);
                    if (dwStatus != 0) pdwStatus = dwStatus;
                    if (hr >= 0)
                    {
                        if (dwStatus == CONNMGR_STATUS_CONNECTED)
                        {
                            bRet = true;
                            break;
                        }
                    }
                }
                Thread.Sleep(100);
            }

            return bRet;
        }

        /// <summary>
        /// 释放所有连接
        /// </summary>
        public void ReleaseConnection()
        {

            if (m_hConnection.ToInt32() != 0)
            {
                ConnMgrReleaseConnection(m_hConnection, 0);
                m_hConnection = IntPtr.Zero;
            }
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct CONNMGR_CONNECTIONINFO
        {
            public uint cbSize;
            public uint dwParams;
            public uint dwFlags;
            public uint dwPriority;
            public int bExclusive;
            public int bDisabled;
            public GUID guidDestNet;
            public IntPtr hWnd;
            public uint uMsg;
            public uint lParam;
            public uint ulMaxCost;
            public uint ulMinRcvBw;
            public uint ulMaxConnLatency;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct CONNMGR_DESTINATION_INFO
        {
            public GUID guid;  // @field GUID associated with network
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = CONNMGR_MAX_DESC)]
            public string szDescription;  // @field Description of network
            public int fSecure; // @field Is it OK to allow multi-homing on the network
        }

        public struct GUID
        {          // size is 16
            public uint Data1;
            public UInt16 Data2;
            public UInt16 Data3;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
            public byte[] Data4;
        }
    }

    /// <summary>
    /// GPRS 連接 Class
    /// </summary>
    public class GPRSmanage
    {
        const int CONNMGR_PARAM_GUIDDESTNET = 0x1; // guidDestNet -  
        const int CONNMGR_PARAM_MAXCOST = 0x2; // ulMaxCost -  
        const int CONNMGR_PARAM_MINRCVBW = 0x4; // ulMinRcvBw -  
        const int CONNMGR_PARAM_MAXCONNLATENCY = 0x8;
        const int CONNMGR_FLAG_PROXY_HTTP = 0x1; // HTTP Proxy
        const int CONNMGR_FLAG_PROXY_WAP = 0x2; // WAP Proxy (gateway)
        const int CONNMGR_FLAG_PROXY_SOCKS4 = 0x4; // SOCKS4 Proxy
        const int CONNMGR_FLAG_PROXY_SOCKS5 = 0x8; // SOCKS5 Proxy
        const int CONNMGR_PRIORITY_VOICE = 0x20000; //  
        const int CONNMGR_PRIORITY_USERINTERACTIVE = 0x08000; //  
        const int CONNMGR_PRIORITY_USERBACKGROUND = 0x02000; //
        const int CONNMGR_PRIORITY_USERIDLE = 0x00800; //
        const int CONNMGR_PRIORITY_HIPRIBKGND = 0x00200; //
        const int CONNMGR_PRIORITY_IDLEBKGND = 0x00080; //
        const int CONNMGR_PRIORITY_EXTERNALINTERACTIVE = 0x00020; //
        const int CONNMGR_PRIORITY_LOWBKGND = 0x00008; //
        const int CONNMGR_PRIORITY_CACHED = 0x00002; //


        const int CONNMGR_STATUS_UNKNOWN = 0x00; //  
        const int CONNMGR_STATUS_CONNECTED = 0x10; // 
        const int CONNMGR_STATUS_DISCONNECTED = 0x20; // 
        const int CONNMGR_STATUS_CONNECTIONFAILED = 0x21; // 
        const int CONNMGR_STATUS_CONNECTIONCANCELED = 0x22; //  
        const int CONNMGR_STATUS_CONNECTIONDISABLED = 0x23; //  
        const int CONNMGR_STATUS_NOPATHTODESTINATION = 0x24; //  
        const int CONNMGR_STATUS_WAITINGFORPATH = 0x25; //  
        const int CONNMGR_STATUS_WAITINGFORPHONE = 0x26; //  
        const int CONNMGR_STATUS_WAITINGCONNECTION = 0x40; //  
        const int CONNMGR_STATUS_WAITINGFORRESOURCE = 0x41; //  
        const int CONNMGR_STATUS_WAITINGFORNETWORK = 0x42; // 
        const int CONNMGR_STATUS_WAITINGDISCONNECTION = 0x80; //  
        const int CONNMGR_STATUS_WAITINGCONNECTIONABORT = 0x81; // 

        private string ConnectUrl;

        unsafe class CONNMGR_CONNECTIONINFO
        {
            public int cbSize = 0x40;
            public int dwParams = 0;
            public int dwFlags = 0;
            public int dwPriority = 0;
            public int bExclusive = 0;
            public int bDisabled = 0;
            public Guid guidDestNet = Guid.Empty;
            public IntPtr hWnd = IntPtr.Zero;
            public int uMsg = 0; // ID Message  
            public int lParam = 0; // lParam  
            public int ulMaxCost = 0;
            public int ulMinRcvBw = 0;
            public int ulMaxConnLatency = 0;
        }

        unsafe class CONNMGR_DESTINATION_INFO
        {
            public Guid guid = Guid.Empty;
            public char* szDescription = null;
        }

        [DllImport("cellcore.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        unsafe static extern int ConnMgrEnumDestinations(int nIndex, CONNMGR_DESTINATION_INFO destinationInfo);

        [DllImport("cellcore.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        unsafe static extern int ConnMgrMapURL(string url, ref Guid guid, int* nIndex);

        [DllImport("cellcore.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        unsafe static extern int ConnMgrEstablishConnection(CONNMGR_CONNECTIONINFO pConnInfo, out IntPtr hConnection);

        [DllImport("cellcore.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        unsafe static extern int ConnMgrEstablishConnectionSync(CONNMGR_CONNECTIONINFO pConnInfo, out IntPtr hConn, uint timeout, uint* status);

        [DllImport("cellcore.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        unsafe static extern int ConnMgrConnectionStatus(IntPtr hConnection, uint* status);

        [DllImport("cellcore.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        unsafe static extern int ConnMgrReleaseConnection(IntPtr hConnection, int bCache);

        [DllImport("coredll.dll")]
        private static extern int CloseHandle(IntPtr hObject);


        public static Hashtable ht = new Hashtable();
        public IntPtr hConnect;

        public static unsafe Guid GuidConnection(int nIndex)
        {
            Guid guid = Guid.Empty;
            CONNMGR_DESTINATION_INFO connMgrDestinationInfo = new CONNMGR_DESTINATION_INFO();
            fixed (char* pDescription = new char[129])
            {
                connMgrDestinationInfo.szDescription = pDescription;
                if (ConnMgrEnumDestinations(nIndex, connMgrDestinationInfo) == 0)
                    guid = new Guid(connMgrDestinationInfo.guid.ToString());
            }
            return guid;
        }

        public static unsafe Guid GuidConnectionURL(string url)
        {
            Guid guid = Guid.Empty;
            int result = ConnMgrMapURL(url, ref guid, null);
            if (result != 0) throw new Exception("ConnMgrMapURL()");
            return guid;
        }

        public static unsafe IntPtr EstablishConnection(Guid guid)
        {
            CONNMGR_CONNECTIONINFO connectionInfo = new CONNMGR_CONNECTIONINFO();
            connectionInfo.dwParams = CONNMGR_PARAM_GUIDDESTNET;
            connectionInfo.dwPriority = CONNMGR_PRIORITY_USERINTERACTIVE;
            connectionInfo.guidDestNet = guid;
            connectionInfo.bExclusive = 1;
            IntPtr hConnection = IntPtr.Zero;
            int result = ConnMgrEstablishConnection(connectionInfo, out hConnection);
            if (result != 0) throw new Exception("ConnMgrEstablishConnection()");
            return hConnection;
        }

        public static unsafe IntPtr EstablishConnection_noConnect(Guid guid)
        {
            CONNMGR_CONNECTIONINFO connectionInfo = new CONNMGR_CONNECTIONINFO();
            connectionInfo.dwParams = CONNMGR_PARAM_GUIDDESTNET;
            connectionInfo.dwPriority = CONNMGR_PRIORITY_USERINTERACTIVE;
            connectionInfo.guidDestNet = guid;
            connectionInfo.bDisabled = 0;
            connectionInfo.bExclusive = 1;
            IntPtr hConnection = IntPtr.Zero;
            int result = ConnMgrEstablishConnection(connectionInfo, out hConnection);
            if (result != 0) throw new Exception("ConnMgrEstablishConnection()");
            return hConnection;
        }

        public static unsafe IntPtr EstablishConnectionSync(Guid guid, double timeout)
        {
            CONNMGR_CONNECTIONINFO connectionInfo = new CONNMGR_CONNECTIONINFO();
            connectionInfo.dwParams = CONNMGR_PARAM_GUIDDESTNET;
            connectionInfo.dwPriority = CONNMGR_PRIORITY_USERINTERACTIVE;
            connectionInfo.guidDestNet = guid;
            IntPtr hConnection = IntPtr.Zero;
            uint tout = (uint)(timeout * 1000 + 1);
            uint status = 0;
            int result = ConnMgrEstablishConnectionSync(connectionInfo, out hConnection, tout, &status);
            if (result != 0) throw new Exception("ConnMgrEstablishConnectionSync()");
            return hConnection;
        }

        public static unsafe void ReleaseConnection(IntPtr hConnection)
        {
            int result = ConnMgrReleaseConnection(hConnection, 0);
            if (result != 0) throw new Exception("ConnMgrReleaseConnection()");
        }

        public bool OpenGprs(string url)
        {
            ConnectUrl = url;

            try
            {
                Guid guid = GuidConnectionURL(ConnectUrl);
                hConnect = EstablishConnection(guid);
            }
            catch (Exception ex)
            {
                return false;
            }
            return true;
        }

        public void CloseGprs()
        {
            try
            {
                Guid guid = GuidConnectionURL(ConnectUrl);
                IntPtr hConnection = EstablishConnection(guid);
                //IntPtr hConnection = EstablishConnection_noConnect(guid);
                ReleaseConnection(hConnection);
                //IsConnected = false;
            }
            catch (Exception ex)
            {
                throw new Exception("GPRS Close Error :" + ex.Message);
            }
        }

        public unsafe uint CheckGprsStatus()
        {
            uint status = 0;

            Guid guid = GuidConnectionURL(ConnectUrl);
            IntPtr hConnection = EstablishConnection_noConnect(guid);
            System.Threading.Thread.Sleep(500);
            int ret = ConnMgrConnectionStatus(hConnection, &status);
            return status;
        }
    }
}