﻿using System;
using System.Runtime.InteropServices;
using System.Threading;
using Microsoft.Win32;
using System.Drawing;

namespace ACLAS
{
    // 44 Bytes
    [StructLayout(LayoutKind.Sequential)]
    public struct CETOUCHINPUT 
    {
        // Specifies the x coordinate of the touch point in 4ths of a pixel.
        public int x;
        
        // Specifies the y coordinate of the touch point in 4ths of a pixel.
        public int y;
        
        // Identifier of the touch input source. Reserved for use by the touch proxy.
        uint hSource;
        
        // Touch point identifier - this is the touch contact index. This ID must be
        // maintained for a contact from the time it goes down until the time it 
        // goes up.
        public uint dwID;
        
        // A set of bit flags that specify various aspects of touch point press / 
        // release and motion.
        public uint dwFlags;
        
        // A set of bit flags that specify which of the optional fields in the 
        // structure contain valid values.
        uint dwMask;
        
        // Time stamp for the event, in milliseconds. If this parameter is 0, 
        // the sample will be timestamped by GWES enroute to the gesture engine 
        // and the TOUCHEVENTMASKF_TIMEFROMSYSTEM flag will be set in dwMask. 
        uint dwTime;
        
        // Specifies the width of the touch contact area in 4ths of a pixel.
        uint cxContact;
        
        // Specifies the height of the touch contact area in 4ths of a pixel.
        uint cyContact;
        
        // Offset to an additional property structure which is associated
        // with this specific touch point. The offset is in bytes from
        // the begining of this CETOUCHINPUT structure.
        uint dwPropertyOffset;
        
        // Size in bytes of the additional property structure.
        uint cbProperty;
    }

    public enum TouchType : uint
    {
        TOUCHTYPE_NOTOUCH = 0, // Denotes the absence of a touch h/w
        // controller
        TOUCHTYPE_SINGLETOUCH = 1, // Denotes a touch h/w controller that can
        // only provide one x/y sample pair at a
        // time, regardless of the number of touch
        // contact points
        TOUCHTYPE_SYMMETRICTOUCH = 2, // Denotes a touch h/w controller that
        // provides scan-curves for the x and y axis
        // separately and hence makes it difficult to
        // accurately pair up the obtained X's and
        // Y's during multi touch sessions
        TOUCHTYPE_MULTITOUCH = 3, // Denotes a h/w controller that is able to
        // distinctly provide x/y pairs for every
        // touch contact point
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct TCH_PROPS
    {
        /* minimum sampling rate supported by the touch driver */
        public uint minSampleRate;
        /* maximum sampling rate supported by the touch driver */
        public uint maxSampleRate;

        /* the touch driver consumer (tchproxy) uses these constants in calibration mode
           to evaluate the user input as a calibration sample*/

        /* minimum number of valid samples before the last down sample is accepted */
        public uint minCalCount;
        /* minimum time in ms for user to hold the pen at calibration cross-hair */
        public uint calHoldSteadyTime;
        /* maximum delta x and delta y user's pen can move from the calibration cross-hair */
        public uint calDeltaReset;

        /* Range of X,Y values for touch sample returned by the driver */
        /* minimum x */
        public int xRangeMin;
        /* maximum x */
        public int xRangeMax;
        /* minimum y */
        public int yRangeMin;
        /* maximum y */
        public int yRangeMax;

        /* Type of touch controller */
        public TouchType touchType;
        /* Touch controller capabilities mask */
        public uint dwTouchInputMask;

        /* Maximum number of simultaneous multitouch samples that the hardware supports 
           ie. Maximum cInputs value passed to the MDD. */
        public uint maxSimultaneousSamples;

        /* Maximum size in bytes of a property associated with a sample. 
           ie. CETOUCHINPUT::cbProperty
           Set to 0 if no extra property data will every be sent with a sample. */
        public uint cbPropertyMax;
    }

    public class TouchEventArgs : EventArgs
    {
        public int X;
        public int Y;
        public int ID;
    }


    public class CusTouch
    {
        public const uint TOUCHEVENTF_MOVE = 0x0001;
        public const uint TOUCHEVENTF_DOWN = 0x0002;
        public const uint TOUCHEVENTF_UP   = 0x0004;
        
        public const uint TOUCHEVENTF_PRIMARY = 0x0010;

        private Device m_CusTouch=null;

        private byte[] m_buf;
        private byte[] m_inbuf = new byte[4];

        private byte[] m_touch_props_buf = new byte[52];

        private TCH_PROPS m_tchprops;

        private Thread thd = null;

        private bool m_close = false;

        private CETOUCHINPUT[] m_tchpoint;

        private bool[] m_Press;

        private int m_Width = 1280;
        private int m_Height = 800;

        public event EventHandler OnMouseDown = null;
        public event EventHandler OnMouseUp = null;
        public event EventHandler OnMouseMove = null;

        public CusTouch()
        {
            Size sz = TVout.GetCusDisplaySize();
            m_Width = sz.Width;
            m_Height = sz.Height; 
        }
        
        public TCH_PROPS TouchInfo
        {
            get
            {
                return m_tchprops;
            }
        }

        public static class IOCTL_TOUCH
        {
            public static readonly uint GET_TOUCH_PROPS = Device.CTL_CODE(FILE_DEVICE.TOUCH, 1, FILE_METHOD.BUFFERED, FILE_ACCESS.ANY);
            public static readonly uint GET_SAMPLE_RATE = Device.CTL_CODE(FILE_DEVICE.TOUCH, 2, FILE_METHOD.BUFFERED, FILE_ACCESS.ANY);
            public static readonly uint SET_SAMPLE_RATE = Device.CTL_CODE(FILE_DEVICE.TOUCH, 3, FILE_METHOD.BUFFERED, FILE_ACCESS.ANY);
            public static readonly uint GET_CALIBRATION_POINTS = Device.CTL_CODE(FILE_DEVICE.TOUCH, 4, FILE_METHOD.BUFFERED, FILE_ACCESS.ANY);
            public static readonly uint SET_SAMPLE_TIMEOUT = Device.CTL_CODE(FILE_DEVICE.TOUCH, 5, FILE_METHOD.BUFFERED, FILE_ACCESS.ANY);
            public static readonly uint ENABLE_TOUCHPANEL = Device.CTL_CODE(FILE_DEVICE.TOUCH, 6, FILE_METHOD.BUFFERED, FILE_ACCESS.ANY);
            public static readonly uint DISABLE_TOUCHPANEL = Device.CTL_CODE(FILE_DEVICE.TOUCH, 7, FILE_METHOD.BUFFERED, FILE_ACCESS.ANY);
            public static readonly uint GET_SAMPLES = Device.CTL_CODE(FILE_DEVICE.TOUCH, 8, FILE_METHOD.BUFFERED, FILE_ACCESS.ANY);
        }

        public bool Open()
        {
            int osize = 0;
            bool rc;

            m_CusTouch = new Device("TCH2:", WINAPI.GENERIC_WRITE | WINAPI.GENERIC_READ);   // 打開驅動裝置類別
            if (!m_CusTouch.Open())
            {
                m_CusTouch = null; 
                return false;
            }

            rc = m_CusTouch.IOControl((int)IOCTL_TOUCH.ENABLE_TOUCHPANEL, null, 0, null,0 , ref osize);

            if (!rc) return false; 
            
            rc = m_CusTouch.IOControl((int)IOCTL_TOUCH.GET_TOUCH_PROPS, null, 0, m_touch_props_buf, m_touch_props_buf.Length, ref osize);

            if (!rc) return false;

            IntPtr intPtr = Marshal.AllocHGlobal(osize);
            Marshal.Copy(m_touch_props_buf , 0, intPtr, osize);
            m_tchprops = (TCH_PROPS) Marshal.PtrToStructure(intPtr, typeof(TCH_PROPS));
            Marshal.FreeHGlobal(intPtr);

            CovTool.UIntToByte(m_tchprops.maxSimultaneousSamples, m_inbuf, 0);
            m_buf = new byte[44 * m_tchprops.maxSimultaneousSamples];

            m_tchpoint = new CETOUCHINPUT[m_tchprops.maxSimultaneousSamples];

            m_Press = new bool[m_tchprops.maxSimultaneousSamples];
                        
            thd = new Thread(TouchProc);
            thd.Priority = ThreadPriority.Highest;

            thd.Start(); 
            return true;
        }

        void TouchProc()
        {
            int n;
            int i;

            while (!m_close)
            {
                n = GetSamples();
                if (n > 0)
                {
                    for (i = 0; i < n; i++)
                    {
                        byte id = (byte)(m_tchpoint[i].dwID);
                        uint flag = m_tchpoint[i].dwFlags;
                        TouchEventArgs args = new TouchEventArgs();
                        args.X = (m_tchpoint[i].x * m_Width) / 10000;
                        args.Y = (m_tchpoint[i].y * m_Height) / 10000;
                        args.ID = id;

                        if ((flag & (TOUCHEVENTF_DOWN | TOUCHEVENTF_PRIMARY)) != 0) // 按下
                        {
                            if (!m_Press[id])
                            {
                                m_Press[id] = true;
                                
                                if (OnMouseDown != null)
                                {
                                    OnMouseDown.Invoke(this, args); 
                                }
 
                            }
                        }
                        else if ((flag & TOUCHEVENTF_UP) != 0) // 放開
                        {
                            if (m_Press[id])
                            {
                                m_Press[id] = false;

                                if (OnMouseUp != null)
                                {
                                    OnMouseUp.Invoke(this, args);
                                }
                            }
                        }
                        else if ((flag & TOUCHEVENTF_MOVE) != 0) // 移動
                        {
                            if (OnMouseMove != null)
                            {
                                OnMouseMove.Invoke(this, args);
                            }

                        }
                    }
                }
                else
                {

                }

            }
        }

        private int GetSamples()
        {
            int n = 0;

            if (m_CusTouch.IOControl((int)IOCTL_TOUCH.GET_SAMPLES, m_inbuf , m_inbuf.Length , m_buf, m_buf.Length, ref n))
            {
                if (n > 0)
                {
                    int size = n * 44;
                    IntPtr intPtr = Marshal.AllocHGlobal(size);
                    Marshal.Copy(m_buf, 0, intPtr, size);

                    IntPtr ptr = intPtr;

                    for (int i = 0; i < n; i++)
                    {
                        m_tchpoint[i] = (CETOUCHINPUT)Marshal.PtrToStructure(ptr, typeof(CETOUCHINPUT));
                        ptr = new IntPtr(ptr.ToInt32() + 44);
                    }
                    Marshal.FreeHGlobal(intPtr);
                }
            }

            return n;
        }

        public void Close()
        {
            int osize = 0;

            m_close = true;

            if (m_CusTouch != null)
            {
                m_CusTouch.IOControl((int)IOCTL_TOUCH.DISABLE_TOUCHPANEL, null, 0, null, 0, ref osize); 
                m_CusTouch.Close();
            }

            if (thd != null)
            {
                thd.Abort(); 
                Thread.Sleep(200);  
            }


        }

    }
}
