﻿using System;
using System.Threading;
using System.Runtime.InteropServices;

namespace ACLAS
{
    public enum DevicePowerState : int
    {

        Unspecified = -1,
        D0 = 0, // Full On: full power, full functionality
        D1, // Low Power On: fully functional at low power/performance
        D2, // Standby: partially powered with automatic wake
        D3, // Sleep: partially powered with device initiated wake
        D4, // Off: unpowered
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct SYSTEM_POWER_STATUS_EX
    {
        public byte ACLineStatus;
        public byte BatteryFlag;
        public byte BatteryLifePercent;
        public byte Reserved1;
        public uint BatteryLifeTime;
        public uint BatteryFullLifeTime;
        public byte Reserved2;
        public byte BackupBatteryFlag;
        public byte BackupBatteryLifePercent;
        public byte Reserved3;
        public uint BackupBatteryLifeTime;
        public uint BackupBatteryFullLifeTime;
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct SYSTEM_POWER_STATUS_EX2
    {
        public byte ACLineStatus;
        public byte BatteryFlag;
        public byte BatteryLifePercent;
        public byte Reserved1;
        public uint BatteryLifeTime;
        public uint BatteryFullLifeTime;
        public byte Reserved2;
        public byte BackupBatteryFlag;
        public byte BackupBatteryLifePercent;
        public byte Reserved3;
        public uint BackupBatteryLifeTime;
        public uint BackupBatteryFullLifeTime;
        public uint BatteryVoltage;
        public uint BatteryCurrent;
        public uint BatteryAverageCurrent;
        public uint BatteryAverageInterval;
        public uint BatterymAHourConsumed;
        public uint BatteryTemperature;
        public uint BackupBatteryVoltage;
        public byte BatteryChemistry;
    }

    public enum ACLineStatus : byte
    {
        Offline = 0,
        Online = 1,
        BackUp = 2,
        Unknown = 255,
    }

    public enum BatteryFlag : byte
    {
        High = 0x01,
        Low = 0x02,
        Critical = 0x04,
        Charging = 0x08,
        NoBattery = 0x80,
        Unknown = 0xFF,
    }

    public enum BatteryChemistry : byte
    {
        Alkaline = 0x01,
        NICD = 0x02,
        NIMH = 0x03,
        LION = 0x04,
        LIPOLY = 0x05,
        ZINCAIR = 0x06,
        Unknown = 0xFF,
    }

    public enum PowerButtonSleepMode : byte
    {
        Default = 0x00 ,
        GotoD1 = 0x01 ,
        GotoD2 = 0x02 ,
        GotoD3 = 0x03
    }


    public class PowerClass
    {
        public const UInt32 POWER_STATE_SUSPEND = 0x00200000;
        public const UInt32 POWER_STATE_ON = 0x00010000;
        public const UInt32 POWER_FORCE = 0x00001000;

        private Device PwrDrawer = null;
        private Thread thdPwr = null;
        private Thread thdKeepWait = null;
        private bool m_wait = false;
        private ManualResetEvent KeepWaitEvent = new ManualResetEvent(false); 
        private PowerType PwrStatus = PowerType.PowerKeyUp;
        private bool thdEnd = false;

        public event EventHandler OnPwrChange = null;

        [DllImport("coredll.dll")]
        public static extern uint GetSystemPowerStatusEx(ref SYSTEM_POWER_STATUS_EX lpSystemPowerStatus,
            bool fUpdate);

        [DllImport("coredll.dll")]
        public static extern uint GetSystemPowerStatusEx2(ref SYSTEM_POWER_STATUS_EX2 lpSystemPowerStatus,
            uint dwLen, bool fUpdate);


        [DllImport("coredll.dll")]
        public static extern UInt32 SetSystemPowerState(
        System.IntPtr psState,
        UInt32  StateFlags,
        UInt32  Options
        );

        [DllImport("coredll.dll")]
        public static extern UInt32 SetSystemPowerState(
        String psState,
        UInt32 StateFlags,
        UInt32 Options
        );

        public PowerClass()
        {
            // 打開驅動裝置類別
            PwrDrawer = new Device( "PKD1:" , WINAPI.GENERIC_WRITE | WINAPI.GENERIC_READ);   
        }

        public Boolean Open()
        {
            Boolean rc =  PwrDrawer.Open();
            if (rc)
            {
                thdKeepWait = new Thread(KeepWaitProc);
                thdKeepWait.Start();

                thdPwr = new Thread(PowerMonitor);
                thdPwr.Start();
            }
            return rc;
        }

        public byte GetCpuTemp()
        {
            byte temp = 0;
            byte[] buf = new byte[1];
            int rsize = 0;
            PwrDrawer.IOControl((int)PKD_IOCTRL.GET_CPU_TEMP, null, 0, buf, 1, ref rsize);
            if (rsize == 1)
            {
                temp = buf[0];
            }
            return temp;
        }

        private void PowerWaitToD3()
        {
            int rsize = 0;
            PwrDrawer.IOControl((int)PKD_IOCTRL.WAIT_TO_D3, null, 0, null, 0, ref rsize);
        }

        private void KeepWaitProc()
        {
            thdKeepWait.Priority = ThreadPriority.Highest;
   
            while(true)
            {
                KeepWaitEvent.WaitOne();
                PowerWaitToD3();
                Thread.Sleep(50);
            }
        }

        public bool KeepWait
        {
            get
            {
                return m_wait;
            }
            set
            {
                m_wait = value;
                if (m_wait)
                {
                    KeepWaitEvent.Set();  
                }
                else
                {
                    KeepWaitEvent.Reset();  
                }
            }
        }

        public PowerType PowerEvent
        {
            get
            {
                return PwrStatus;
            }
        }

        public void Close()
        {
            bool wait = false;

            thdEnd = true;

            if (thdPwr != null)
            {
                thdPwr.Abort();
                wait = true;
            }

            if (thdKeepWait != null)
            {
                thdKeepWait.Abort();
                wait = true;
            }

            if (wait)
            {
                Thread.Sleep(100);  
            }

            PwrDrawer.Close();  
        }

        private void PowerMonitor()
        {
            byte[] obuf = new byte[1];
            int rsize;
            while (!thdEnd)
            {
                rsize = 0;
                PwrDrawer.IOControl((int)PKD_IOCTRL.MONITOR_POWER, null, 0, obuf, 1, ref rsize);
                PwrStatus = (PowerType)obuf[0];

                if (OnPwrChange != null)
                {
                    OnPwrChange.Invoke(this, null);
                }
            }
        }

        public bool GetDeviceID(ref string version, ref string id)
        {
            id = "";
            version = "";
            byte[] buf = new byte[13];
            int rsize = 0;
            PwrDrawer.IOControl((int)PKD_IOCTRL.GET_PRODUCT_ID, null, 0, buf, 13, ref rsize);

            if (rsize == 13)
            {
                version = (buf[0] >> 4).ToString() + "." + (buf[0] & 0xf).ToString();
                for (int i = 1; i < 13; i++)
                    id += buf[i].ToString("X2");

                return true;
            }
            return false;
        }

        public void OpenDraw(int n)
        {
            byte[] buf = { (byte)('0'+n) };
            int rsize = 0;
            PwrDrawer.IOControl((int)PKD_IOCTRL.OPEN_DRAWER, buf, 1, null, 0, ref rsize);
        }

        public void SetPowerButtonSleepMode(PowerButtonSleepMode mode)
        {
            byte[] buf = new byte[1];
            int rsize = 0;
            buf[0] = (byte) mode;
            PwrDrawer.IOControl((int)PKD_IOCTRL.SET_M3_UPS_MODE , buf, 1, null, 0, ref rsize);
        }

        public bool SetPowerWaitToD3Mode(bool on)
        {
            byte[] buf = new byte[1];
            int rsize = 0;
            buf[0] = (byte)(on ? 1 : 0);
            return PwrDrawer.IOControl((int)PKD_IOCTRL.SET_WAIT_TO_D3,buf, 1, null, 0, ref rsize);
        }

        public void GPRSPowerControlOn()
        {
            byte[] buf ={1};
            int rsize = 0;
            PwrDrawer.IOControl((int)PKD_IOCTRL.SET_GPRS_POWER_CONTROL , buf, 1, null, 0, ref rsize);
        }

        public void GPRSPowerControlOff()
        {
            byte[] buf = { 0 };
            int rsize = 0;
            PwrDrawer.IOControl((int)PKD_IOCTRL.SET_GPRS_POWER_CONTROL, buf, 1, null, 0, ref rsize);
        }

        // turn off power
        public void ShutdownPower()
        {
            PwrDrawer.ShutdownPower();
        }

        // goto to sleep mode 
        public void SleepPower()
        {
            SetSystemPowerState(IntPtr.Zero, POWER_STATE_SUSPEND, POWER_FORCE); 
        }

        // Goto to System Idel (turn off screen) mode
        public void D2Power()
        {
            SetSystemPowerState("systemidle" , POWER_STATE_SUSPEND, POWER_FORCE);
        }

        // Goto Power All On Mode
        public static void D0Power()
        {
            SetSystemPowerState(IntPtr.Zero, POWER_STATE_ON , POWER_FORCE);
        }

        public uint GetCpuSpeed()
        {
            uint speed = 0;
            byte[] buf = new byte[4];
            int rsize = 0;
            PwrDrawer.IOControl((int)PKD_IOCTRL.GET_CPU_SPEED, null, 0, buf, 4, ref rsize);
            if (rsize == 4)
            {
                speed = CovTool.toUInt(buf, 0);
            }
            return speed;
        }

        public int GetCpuIdle(byte[] buf)
        {
            int rsize = 0;
            PwrDrawer.IOControl((int)PKD_IOCTRL.GET_CPU_IDLE, null, 0, buf, 4, ref rsize);
            return rsize;
        }

    }
}
