﻿using System;
using System.Runtime.InteropServices;
using System.Threading;
using System.Drawing;
using Microsoft.Win32;

namespace ACLAS.Printer
{
    public enum PRN_IOCTRL
    {
        SETSTROBETIMER = 0,
        FEED = 1,
        PRN_IOC_REWIND = 2,
        PRN_IOC_STOP = 3,
        GETDOTWIDTH = 4,
        GETSPAREEREA = 5,
        GETUNPRINTLEN = 6,
        GETSTATUS = 7,
        CLRAllBUF = 8,
        SETAREACONSTRAST = 9,
        REALPRINTFLAG = 10,
        FEEDSTOP = 11,
        PAPER_STATUS = 12,        // 缺紙狀態
        PRINTER_CONTINUE = 17,    // 樟哿湖荂
        PAPER_CHK_SET = 18,       // 萸淝潰脤潔路扢离
        PAPER_FEED_LEN_SET = 19   // 輛祧酗僅扢离
    }

    public enum PrnStatus
    {
        Stop, Ready, Start, PrintNow, Done, Break
    }

    public static class AclasPrinter
    {
        private const String DllName = "tvout.dll";
        [DllImport(DllName)]
        private static extern void DeInit();
        [DllImport(DllName)]
        private static extern Boolean InitPrinter(uint handle , int width);
        [DllImport(DllName)]
        private static extern Boolean PrintImage(IntPtr hbm, int Width, int Height);
        
        private static PrnStatus PrnStep = PrnStatus.Stop;

        private static Bitmap PrnBitMap = null;
        private static int PrnHeight = 0;

        private static Thread PrnThd = null;

        private static Boolean Cutter = false;

        private static Boolean ChkCutter = false;

        private static Device Printer = null;

        private static int m_PrnWidth; // 打印機寬度 點

        private static AutoResetEvent eventWaitStart = new AutoResetEvent(false); 

        public static int Width
        {
            get
            {
                return m_PrnWidth;
            }
        }

        public static Boolean HaveCutter
        {
            get
            {
                if (!ChkCutter)
                {
                    try
                    {
                        RegistryKey PrnKey = Registry.LocalMachine.OpenSubKey(@"Drivers\BuiltIn\AclasPNT", false);
                        UInt32 cut = Convert.ToUInt32(PrnKey.GetValue("Cutter"));
                        Cutter = (cut != 0);
                    }
                    catch 
                    {
                    }
                    ChkCutter = true;
                }
                return Cutter;
            }
        }

        public static bool Open()
        {
            Byte[] buf = new Byte[4];
            int rsize = 0;
            Printer = new Device("PNT1:", WINAPI.GENERIC_WRITE | WINAPI.GENERIC_READ);   // 打開驅動裝置類別

            if (!Printer.Open()) return false;

            if (Printer.IOControl((int)PRN_IOCTRL.GETDOTWIDTH, null, 0, buf, 4, ref rsize))
            {
                m_PrnWidth = CovTool.toInt(buf, 0);
            }
            else
            {
                m_PrnWidth = 384;
            }
            if (!AclasPrinter.InitPrinter((uint) Printer.Handle , m_PrnWidth ))
                return false;
            PrnThd = new Thread(PrnOrder);
            PrnThd.Start();
            return true;
        }

        /// <summary>
        /// Alloc a New Bitmap page to print
        /// </summary>
        /// <param name="Height">Page Height by Dot</param>
        /// <returns>Returns print page Bitmap</returns>
        public static Bitmap SetPageSize(int Height)
        {
            if (PrnBitMap != null)
            {
                PrnBitMap.Dispose();
            }
            PrnBitMap = new Bitmap(m_PrnWidth , Height, System.Drawing.Imaging.PixelFormat.Format16bppRgb565);
            return PrnBitMap;
        }

        public static void Close()
        {
            if (PrnThd != null)
                PrnThd.Abort();
            Thread.Sleep(100);

            if (PrnBitMap != null)
            {
                PrnBitMap.Dispose();
            }
            AclasPrinter.DeInit();
        }

        /// <summary>取得印表機狀態</summary> 
        /// <returns>FALSE 列印中  TURE 列印完成了</returns> 
        public static Boolean GetPrnStatus()
        {
            Byte[] buf = new Byte[4];
            int rsize = 0;
            Printer.IOControl((int)PRN_IOCTRL.GETSTATUS, buf, 4, null, 0, ref rsize);
            return (buf[0] == 0);
        }

        /// <summary>取得缺紙狀態</summary> 
        /// <returns>TRUE 缺紙  FALSE 不缺紙</returns> 
        public static Boolean GetPaperOutStatus()
        {
            Byte[] buf = new Byte[4];
            int rsize = 0;
            Printer.IOControl((int)PRN_IOCTRL.PAPER_STATUS , buf, 4, null, 0, ref rsize);
            return (buf[0] != 0);
        }

        private static Boolean Feed(int feed)
        {
            int osize = 0;
            Byte[] buf = new Byte[4];
            Byte[] obuf = null;
            CovTool.IntToByte(feed, buf, 0);
            return Printer.IOControl((int)PRN_IOCTRL.FEED, buf, buf.Length, obuf, 0, ref osize);
        }

        public static Boolean LineFeed(int feed)
        {
            feed = (int)(((uint)feed) & 0x7fffffff);
            return Feed(feed);
        }

        public static Boolean BackLineFeed(int feed)
        {
            feed = -feed;
            feed = (int)(((uint)feed) | 0x80000000);
            return Feed(feed);
        }

        public static bool Write(Byte[] buf, int len)
        {
            return Printer.Write(buf, len);
        }

        public static bool PrintString(string s)
        {
            Byte[] buf = CovTool.Str2Byte(s);
            return Printer.Write(buf, buf.Length);
        }

        public static void BmpToPrinter(int Height)
        {
            PrnHeight = Height;
            IntPtr hbm = PrnBitMap.GetHbitmap();
            AclasPrinter.PrintImage(hbm, m_PrnWidth, PrnHeight);
            WINAPI.DeleteObject(hbm);
        }

        public static void Start()
        {
            PrnStep = PrnStatus.Start;
            eventWaitStart.Set();
        }

        public static PrnStatus GetStatus()
        {
            return PrnStep;
        }

        public static void FullCutWithFeed(int line)
        {
            Byte[] cut = new Byte[] { 0x1d, 0x56, 0x41, 0 };
            cut[3] = (Byte)line;
            Write(cut, cut.Length);
        }

        public static void HalfCutWithFeed(int line)
        {
            Byte[] cut = new Byte[] { 0x1d, 0x56, 0x42, 0 };
            cut[3] = (Byte)line;
            Write(cut, cut.Length);
        }

        public static void FullCut()
        {
            Byte[] cut = new Byte[] { 0x1d, 0x56, 0x30 };
            Write(cut, cut.Length);
        }

        public static void HalfCut()
        {
            Byte[] cut = new Byte[] { 0x1d, 0x56, 0x31 };
            Write(cut, cut.Length);
        }

        // 設定打印濃度  -9 (最淡) 到 15 (最濃)  0 = 出廠建議濃度
        // 注意濃度值越高 加熱時間越久 打印速度越慢 而且有可能會降低打印頭的使用期限
        public static void Shade(int ct)
        {
            int osize = 0;
            Byte[] buf = new Byte[4];
            CovTool.IntToByte(ct , buf, 0);
            Printer.IOControl((int)PRN_IOCTRL.SETAREACONSTRAST, buf, buf.Length, null, 0, ref osize);
        }

        // 打印執行緒
        private static void PrnOrder()
        {
            int WaitCnt;

            while (true)
            {
                eventWaitStart.WaitOne(); 

                if (PrnHeight != 0)
                {
                    PrnStep = PrnStatus.PrintNow;

                    WaitCnt = 4;
                    while (PrnStep == PrnStatus.PrintNow)
                    {
                        if (GetPrnStatus())
                        {
                            WaitCnt--;
                            if (WaitCnt == 0)
                            {
                                PrnStep = PrnStatus.Done;
                                break;
                            }
                        }
                        else
                        {
                            WaitCnt = 4;
                        }
                        Thread.Sleep(50);
                    }
                }
                else
                {
                    PrnStep = PrnStatus.Done;
                }
            }
        }
    }
}
