﻿using System;
using System.Threading;
using System.Windows.Forms;
using ACLAS;

namespace ACLAS.RFID
{
    public class RFIDDriver
    {
        public enum CardResult
        {
            OK , BAD , ValueBlockBad , TimeOut
        }

        public enum BlockReadMode
        {
            Value, Data
        }

        public enum AccessRight
        {
            Never, KeyA, KeyB, KeyAorB
        }

        // 空白卡片的 Key A = 0xffffffffffff 6 個 0xff
        public readonly  Byte[] InitKey = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

        public readonly AccessRight[,] RightTbl = new AccessRight[8, 4] 
        {
            { AccessRight.KeyAorB , AccessRight.KeyAorB , AccessRight.KeyAorB, AccessRight.KeyAorB } ,
            { AccessRight.KeyAorB , AccessRight.KeyB    , AccessRight.Never  , AccessRight.Never   } ,
            { AccessRight.KeyAorB , AccessRight.Never   , AccessRight.Never  , AccessRight.Never   } ,
            { AccessRight.KeyAorB , AccessRight.KeyB    , AccessRight.KeyB   , AccessRight.KeyAorB } ,
            { AccessRight.KeyAorB , AccessRight.Never   , AccessRight.Never  , AccessRight.KeyAorB } ,
            { AccessRight.KeyB    , AccessRight.Never   , AccessRight.Never  , AccessRight.Never   } ,
            { AccessRight.KeyB    , AccessRight.KeyB    , AccessRight.Never  , AccessRight.Never   } ,
            { AccessRight.Never   , AccessRight.Never   , AccessRight.Never  , AccessRight.Never   } ,
        };

        public enum JobStatus
        {
            Done,              // 工作完成了
            Bad,               // 強制回到 IDLE MODE 結束工作 
            Start,             // 卡片操作開始
            WaitCard,          // 等待卡片
            DetectCard,        // 偵測到卡片    
            Idle,              // 閒置
            WaitCardLeave ,    // 等待卡片移開
            Auth,              // 授權
            AuthOK,            // 授權成功
            Transfer,          // 加減值後存入BLOCK 中
            GetValue,          // 讀取卡片目前的金額 Value    
            GetRight           // 讀取卡片區塊權限
        }

        public enum JobMode
        {
            GetCardNo ,   // 讀取卡號
            ReadBlock ,   // 讀取區塊
            WriteBlock,   // 寫入區塊
            InitValue,    // 啟始儲值區塊
            IncValue ,    // 加值
            DecValue,     // 扣值
            AccessMode    // 改變存取權限
        }

        public enum KeyType
        {
            KeyA = 0 ,
            KeyB = 4
        }

        private Boolean DebugMode = false;

        private Byte SectorNo;  // Sector 號碼  0 - 15
        private Byte BlockNo;   // Block  號碼  0 - 3
        private int  BlockAddr; // 區塊地址
        private KeyType keytype = KeyType.KeyA;
        private Byte[] key;  // 使用的密鑰

        private const int DataLen = 26;
        private Byte[] inBuf = new Byte[DataLen];
        private Byte[] OutBuf = new Byte[DataLen];
        private int rsize = 0;
        private Device RfidCard = new Device("RID1:", 0);   // 打開驅動裝置類別
      
        Thread  thdRFID = null;
        private JobStatus WorkStep = JobStatus.Idle;
        private JobMode JobStep;
        private JobMode Job;  
        private Byte[] TagType = null;
        private String CardNo; // 卡號
        private String CardType;
        private Byte[] SNR = null;
        private Byte CardSize;
        private bool IsOpen = false;

        private bool LoopMode = false;

        private Byte[] WriteBuffer = new Byte[16];
        private UInt32 value;

        public  int MaxOperationTime = 2000;
        private Boolean TimeOut;

        private System.Threading.Timer tmr;     

        public  event EventHandler OnReadDone = null;

        public UInt16[] SoundOK = new UInt16[] {
            1 , 2600, 100 , 40 ,
        };

        public UInt16[] RFIDReady = new UInt16[] {
            2 , 2000, 60 , 40 ,
            1 , 3000, 200 , 40 
        };

        public UInt16[] SoundBad = new UInt16[] {
            3 , 4000, 100 , 40 
        };

        private TextBox tbLog = null;

        public  BlockReadMode ReadMode;

        private int p_NewValue;

        // 取得卡片目前的金額
        private Boolean GetOldValue;
        // 卡片目前的值
        private int p_OldValue;

        private CardResult p_CardResult;

        private Boolean p_GetAccess;

        private Byte[] AccessBlock = new Byte[16];

        private Byte AccessMode;

        private bool UseDefaultKey;

        public void SetAccessBlockKey(Byte[] a , Byte[] b , Byte mode , bool dk)
        {
            Array.Copy(a,0,  AccessBlock, 0,6);
            Array.Copy(b, 0,  AccessBlock, 10,6);
            AccessMode = mode;
            UseDefaultKey = dk;
        }

        public int OldValue
        {
            get
            {
                return p_OldValue;
            }
        }

        public int NewValue
        {
            get
            {
                return p_NewValue;
            }
        }

        public CardResult Result
        {
            get
            {
                return p_CardResult; 
            }
        }

        public RFIDDriver(TextBox tb)
        {
            tbLog = tb;
            ShowLog(
                    "Aclas RFID Card Demo v1.000a\r\n" +
                    "Copyright © 2012 all right reserved\r\n" +
                    "Designed by Jacky Lin\r\n" 
                  );

             TimerCallback tcb = new TimerCallback(TimeOutProc);
             tmr = new System.Threading.Timer(tcb, null, Timeout.Infinite , Timeout.Infinite);
        }

        public void SetLoopMode(Boolean loop)
        {
            LoopMode = loop;
        }

        void  TimeOutProc(Object state)
        {
            tmr.Change(Timeout.Infinite, Timeout.Infinite); 
            TimeOut = true;
        }

        public void SetBlockReadMode(BlockReadMode mode)
        {
            ReadMode = mode;
        }

        public void SetDebugMode(bool on)
        {
            DebugMode = on;
        }

        public String GetHexData()
        {
            return CovTool.HexString(OutBuf, 3, OutBuf[2]);   
        }

        public Byte[] GetRead()
        {
            Byte[] buf = new Byte[16];
            if (OutBuf[2] != 16)
                return null;
            Array.Copy(OutBuf, 3, buf, 0, 16);
            return buf;
        }

        public void PutValue(UInt32 val)
        {
            value = val;
        }

        public bool PutWriteData(Byte[] buf)
        {
            if (buf.Length != 16)
            {
                ShowLog("Error! Put data to buffer length <> 16\r\n");
                return false;
            }
            Array.Copy(buf, WriteBuffer, 16);   
            return true;
        }

        public Boolean Open()
        {
            IsOpen = RfidCard.Open();
            return IsOpen;
        }

        public bool SelectKey(KeyType kt , Byte [] k)
        {
            keytype = kt;
            key = k;
            return (key.Length == 6);  // Key 長度必需為 6
        }

        public bool SelectPos(Byte s , Byte b)
        {
            if (s<0 || s>15)
                return false;
            if (b<0 || b > 3)
                return false;
            if (s==0 && b==0)
                return false;

            SectorNo = s;
            BlockNo = b;

            BlockAddr = s * 4 + b;
            return true;
        }

        public bool DoJob(JobStatus mode,JobMode type)
        {
            if (WorkStep != JobStatus.Idle)
            {
                ShowDebug("Kill Old RFID Job\r\n");
                Abort();
            }  

            thdRFID = new Thread(RFIDMain);
            WorkStep = mode;
            Job = JobStep = type;
            thdRFID.Start();
            return true;
        }

        private void Abort()
        {
            if (thdRFID != null)
            {
                thdRFID.Abort();
                Thread.Sleep(100);
                WorkStep = JobStatus.Idle;
                thdRFID = null;
                TurnOff();   // Turn off RDIF 
                ShowLog("RFID.Abort()\r\n"); 
            }
        }

        public void Close()
        {
            if (IsOpen)
            {
                Abort();
                RfidCard.Close();
            }
        }

        // 偵測到卡片 進行卡片工作
        private void DetectCard()
        {
            switch (JobStep)
            {
                case JobMode.AccessMode: // 權限變更 
                case JobMode.IncValue:   // 加值
                case JobMode.DecValue:   // 扣值 
                case JobMode.WriteBlock: // 寫入區塊 
                case JobMode.ReadBlock:  // 讀取區塊
                    WorkStep = JobStatus.Auth; 
                    break;
                default:
                    // 不正確的工作型態
                    WorkStep = JobStatus.Bad; 
                    break;
            }
        }

        private bool CardAuth()
        {
            RetCode rc = Mifs_AuthKey(keytype, BlockAddr, key);
            
            if (rc != RetCode.MI_OK)
            {
                ShowDebug ( "Card Authentication fail!\r\nreturn code : " + rc.ToString()+"\r\n");
                return false;
            }
            ShowDebug("Card authentication OK!\r\n");
            return true;
        }

        //卡片操作
        private bool CardOperation()
        {
            RetCode r;
            bool rc = false; 

            switch (JobStep)
            {
                case JobMode.AccessMode:
                    if (!p_GetAccess)
                    {
                        WorkStep = JobStatus.GetRight;
                        return true;
                    }
                    break;
                case JobMode.IncValue:
                    if (!GetOldValue)
                    {
                        WorkStep = JobStatus.GetValue;  
                        return true;
                    }

                    r = Mifs_IncValue(BlockAddr);
                    if (r != RetCode.MI_OK)
                    {
                        ShowDebug("Inc Value Error! return code " + r.ToString() + "\r\n");
                        return false;
                    }
                    ShowDebug("Card Inc Value command done!\r\n");
                    WorkStep = JobStatus.Transfer;
                    return true;

                case JobMode.DecValue:

                    if (!GetOldValue)
                    {
                        WorkStep = JobStatus.GetValue;
                        return true;
                    }

                    r = Mifs_DecValue(BlockAddr);
                    if (r != RetCode.MI_OK)
                    {
                        ShowDebug("Dec Value Error! return code " + r.ToString() + "\r\n");
                        return false;
                    }
                    ShowDebug("Card Dec Value command done!\r\n");
                    WorkStep = JobStatus.Transfer;
                    return true;

                case JobMode.WriteBlock:
                    r = Mifs_Write(BlockAddr);
                    if (r != RetCode.MI_OK)
                    {
                        ShowDebug("Write Error! return code " + r.ToString() + "\r\n");
                        return false;
                    }
                    ShowLog("Card Write OK!\r\n");
                    WorkStep = JobStatus.Done;  
                    if (OnReadDone != null)
                        OnReadDone.Invoke(this, null);
                    return true;

                case JobMode.ReadBlock:
                    rc = ReadCard(ref p_NewValue);
                    if (rc)
                    {
                        // ShowLog("Read Value is "+  p_NewValue + "\r\n"); 
                        ShowLog("Card Read OK! Sector:"+ SectorNo +" Block:"+ BlockNo +"\r\n");
                        WorkStep = JobStatus.Done;
                        if (OnReadDone != null)
                            OnReadDone.Invoke(this, null);
                    }
                    return rc;
                default:
                    break;
            }

            return rc;
        }

        public void ChangeWorkMode(JobStatus js)
        {
            WorkStep = js;
        }

        private bool ReadCard(ref int val)
        {
            RetCode r = Mifs_Read(BlockAddr);

            if (r != RetCode.MI_OK)
            {
                ShowDebug("Read Error! return code " + r.ToString() + "\r\n");
                return false;
            }

            if (ReadMode == BlockReadMode.Value)
            {
                Byte[] buf = GetRead();
                if (buf != null)
                {
                    val = 0;
                    if (CovTool.GetValue(SectorNo, BlockNo, ref val , buf))
                    {
                        ShowDebug("Read value is " + val + "\r\n");
                    }
                    else
                        buf = null;
                }

                if (buf == null)
                {
                    ShowLog("Read Value Block Error!\r\n" +
                        "This Block is not Value Block,\r\n" +
                            "or this Value Block is bad\r\n");

                    p_CardResult = CardResult.ValueBlockBad;  

                }
            }
            return true; 
        }

        // 產生要寫入的權限資料
        private void MakeAccessBlock(   )
        {
            Byte[] buf = new Byte[4];

            Array.Copy(OutBuf, 9 , buf, 0 , 4);

            Byte[] Cb = new Byte[4];

            Byte mask =(Byte) (0x10 << (BlockNo & 3));

            if ((AccessMode & 1) == 0) // bit 1
            {
                buf[1] = (Byte)(buf[1] & (~mask));
            }
            else
            {
                buf[1] = (Byte)(buf[1] | mask);
            }

            if ((AccessMode & 4) == 0) // bit 3
            {
                buf[2] = (Byte)(buf[2] & (~mask));
            }
            else
            {
                buf[2] = (Byte)(buf[2] | mask);
            }

            mask = (Byte) (mask >> 4);

            if ((AccessMode & 2) == 0) // bit 2
            {
                buf[2] = (Byte)(buf[2] & (~mask));
            }
            else
            {
                buf[2] = (Byte)(buf[2] | mask);
            }

            buf[2] = (Byte)(buf[2] | 0x88); // C3.3 = 1 &  C2.3 = 1

            buf[1] = (Byte)(buf[1] & (~0x80));  // C1.3 = 0

            if (UseDefaultKey)
            {
                buf[2] = (Byte)(buf[2] & (~0x08)); // C3.2 = 0
            }

            buf[0]  = (Byte)( ((~buf[2]) << 4) | ((buf[1] ^ 0xf0 ) >> 4) );
            buf[1]  = (Byte)( (buf[1] & 0xf0)  | ((buf[2] ^ 0xf0 ) >> 4) );

            Array.Copy(buf, 0,  AccessBlock , 6, 4);
        }

        private bool ReadRight()
        {
            BlockAddr |= 3;
            RetCode r = Mifs_Read(BlockAddr | 3);

            if (r != RetCode.MI_OK)
            {
                ShowDebug("Read Error! return code " + r.ToString() + "\r\n");
                return false;
            }
            p_GetAccess = true;

            MakeAccessBlock();

            JobStep = JobMode.WriteBlock;   // 把權限寫入

            return true;
        }

        private bool TransferCard()
        {
            RetCode r;
            r = Mifs_Transfer(BlockAddr);
            if (r != RetCode.MI_OK)
            {
                ShowDebug("Card Transfer Error! return code " + r.ToString() + "\r\n");
                return false;
            }
            ShowLog("Card Transfer OK!\r\n");
            JobStep = JobMode.ReadBlock;  
            WorkStep = JobStatus.AuthOK;
            return true;
        }

        private void RFIDMain()
        {
            Boolean Abort = false;
            TimeOut = false;
            Boolean Done = false;

            while (!Abort)
            {
                if (TimeOut)
                {
                    if (Done)
                    {
                        p_CardResult = CardResult.OK;
                        ShowLog("Card access Ok!\r\n");
                        Buzzer.SoundOut(SoundOK);
                    }
                    else
                    {
                        p_CardResult = CardResult.TimeOut;
                        ShowLog("Card access time out!\r\n");
                        Buzzer.SoundOut(SoundBad);
                    }

                    if (LoopMode)
                        WorkStep = JobStatus.Start;
                    else
                        break;
                }

                switch (WorkStep)
                {
                    case JobStatus.Done:
                        Buzzer.SoundOut(SoundOK);
                        if (LoopMode)
                            WorkStep = JobStatus.WaitCardLeave; 
                        else
                            Abort = true;
                        break;
                    case JobStatus.Bad:
                        if (Done)
                        {
                            WorkStep = JobStatus.Done;
                        }
                        else
                        {
                            ShowLog("Card Access Error!\r\n");
                            Buzzer.SoundOut(SoundBad);
                            p_CardResult = CardResult.BAD;

                            goto loop;
                        }
                        break;
                    case JobStatus.Start:
                        CardNo = "";
                        ShowLog("Job Mode : " + JobStep.ToString() + "\r\n" +
                                  "Card authentication use " + keytype.ToString() + "\r\n" +
                                  "Access Sector:" + SectorNo + " Block:" + BlockNo + "\r\n" +
                                  keytype.ToString() + ": [" + CovTool.HexString(key, 0, 6) + "]\r\n" +
                                  "Waiting for Card....\r\n"
                               );
                        tmr.Change(Timeout.Infinite, Timeout.Infinite);
                        TimeOut = false;
                        Buzzer.SoundOut(RFIDReady);
                        WorkStep = JobStatus.WaitCard;
                        Done = false;
                        GetOldValue = false;
                        p_GetAccess = false;
                        break;

                    case JobStatus.WaitCard:

                        if (ReadCardNoA() == RetCode.MI_OK)
                        {
                            if (JobStep == JobMode.GetCardNo)
                            {
                                WorkStep = JobStatus.WaitCardLeave;
                                Buzzer.SoundOut(SoundOK);
                            }
                            else
                            {
                                WorkStep = JobStatus.DetectCard;
                            }
                        }
                        break;

                    case JobStatus.DetectCard:  // 偵測到卡片了

                        RetCode rc = mifs_select(); // 選擇卡片

                        if (rc != RetCode.MI_OK)
                        {
                            ShowDebug("Select Card Fail\r\n");
                            WorkStep = JobStatus.WaitCard;
                        }
                        ShowDebug("Select Card OK!\r\n");
                        DetectCard();
                        break;

                    case JobStatus.Auth:
                        if (CardAuth()) // 卡片授權成功
                        {
                            WorkStep = JobStatus.AuthOK;
                        }
                        else
                        {
                           ShowDebug("Waiting for Card\r\n");
                           WorkStep = JobStatus.WaitCard;
                           // ShowDebug("Select for Card again\r\n");
                           // WorkStep = JobStatus.DetectCard;
                        }
                        break;
                    case JobStatus.AuthOK:
                        if (!CardOperation())
                        {
                            ShowDebug("Card Operation Again!\r\n");
                            WorkStep = JobStatus.WaitCard;
                        }
                        break;
                    case JobStatus.GetValue:
                        if (ReadCard(ref p_OldValue))
                        {
                            ShowLog("Card old value is " + p_OldValue + "\r\n");
                            GetOldValue = true;
                            WorkStep = JobStatus.AuthOK;
                        }
                        else
                            WorkStep = JobStatus.Auth;
                        break;
                    case JobStatus.GetRight:
                        if (ReadRight())
                        {
                            WorkStep = JobStatus.AuthOK;
                        }
                        else
                            WorkStep = JobStatus.Auth;
                        break;
                    case JobStatus.Transfer:
                        if (!TransferCard())
                        {
                            ShowDebug("Card Transfer Again!\r\n");
                            WorkStep = JobStatus.AuthOK;
                        }
                        else
                        {
                            Done = true;

                            switch (Job)
                            {
                                case JobMode.DecValue:
                                    p_NewValue = (int)(p_OldValue - value);
                                    break;
                                case JobMode.IncValue:
                                    p_NewValue = (int)(p_OldValue + value);
                                    break;
                            }
                            ShowLog("Transfer Value is from " + p_OldValue + " to " + p_NewValue + "\r\n");
                            if (OnReadDone != null)
                                OnReadDone.Invoke(this, null);
                        }
                        break;

                    case JobStatus.WaitCardLeave:
                        // 連續 5 次都找不到卡片 , 以確定卡片已經移出感應區 , 以防止連續操作卡片. 例如重覆扣卡
                        for (int i = 0; i < 5; i++)
                        {
                            if (ReadCardNoA() == RetCode.MI_OK)
                                i = 0;
                        }
                        ShowLog("RFID Card Leave!\r\n");

                        TimeOut = false;

                        if (LoopMode)
                        {
                            WorkStep = JobStatus.Start;  
                        }
                        else
                            Abort = true;
                        break;
                    default:
                        ShowLog("Unknow Work Status : " + ((int)WorkStep) + "\r\n");
                    loop:
                        if (LoopMode)
                            WorkStep = JobStatus.Start;
                        else
                            Abort = true;
                    break;
                }
            }

            TurnOff();   // Turn off RDIF 
            ShowLog("RFID Thread End\r\n"); 
            WorkStep = JobStatus.Idle;
            thdRFID = null;
        }

        // inBuf and OutBuf Struct Toatl 26 Bytes
        // 0 BYTE pkg_no;
        // 1 BYTE cmd;          -> Command
        // 2 BYTE len;          -> Length  
        // 3 BYTE dat[22];      -> Data
        // BYTE bcc;

        public RetCode Command(CommandCode code, Byte len, Byte[] buf)
        {
            inBuf[0] = 0;  // reserve 
            inBuf[1] = (Byte)code;
            inBuf[2] = len;
            if (len > 22) return RetCode.MI_BYTECOUNTERR;  // input data length too big
            for (int i = 0; i < len; i++)
                inBuf[i + 3] = buf[i];
            rsize = 0;
            RfidCard.IOControl((int) IOCtrl.ReadCard, inBuf, DataLen, OutBuf, DataLen, ref rsize);
            RetCode rc = (RetCode)OutBuf[1];
            return rc;
        }

        public void  TurnOff()
        {
            rsize = 0;
            RfidCard.IOControl((int)IOCtrl.Turn_OFF , null ,  0 , null , 0, ref rsize);
        }

        public Boolean  GetInfo(ref int SerialNo,ref String ProductClassCode,ref String DriverVersion)
        {
            Boolean rc;

            rc = (mifs_set_config() == RetCode.MI_OK);

            if (rc)
            {
                rc = (Command(CommandCode.GetInfo, 0, null) == RetCode.MI_OK);
                if (rc)
                {
                    ProductClassCode = CovTool.HexString(OutBuf, 3, 5);
                    SerialNo = CovTool.toInt(OutBuf, 8);
                    DriverVersion = "Ver:" + (OutBuf[12] >> 4) + "." + (OutBuf[12] & 0xf);
                }
            }
            return rc;
        }

        delegate void TextCallBack(String msg);

        public void ShowLog(String msg)
        {
            if (tbLog == null) return; 

            if (tbLog.InvokeRequired)
            {
                TextCallBack tb = new TextCallBack(ShowLog);
                tbLog.Invoke(tb, new Object[] { msg });
            }
            else
            {
                tbLog.Select(tbLog.TextLength, 0);
                tbLog.SelectedText = msg;
                tbLog.Select(tbLog.TextLength, 0);
                tbLog.ScrollToCaret();
            }
        }

        public void ShowDebug(String msg)
        {
            if (!DebugMode) return;
            ShowLog(msg);
        }

        private RetCode mifs_set_config()
        {
            return Command(CommandCode.Config,0,null);
        }

        private RetCode mifs_Anticoll2(Byte Bcnt)
        {
            RetCode rc;
            Byte[] buf = new Byte[1];
            buf[0] = Bcnt;
            rc = Command(CommandCode.Anticoll2, 1, buf);
            if (rc == RetCode.MI_OK)
            {
                int len = OutBuf[2];
                if (len < 4) len = 4;
                SNR = new Byte[len]; 
                for (int i=0;i<len;i++)
                    SNR[i] = OutBuf[3+i];
            }
            return rc;
        }

        private RetCode mifs_request(Byte mode)
        {
            RetCode rc;
            Byte[] buf = new Byte[1];

            buf[0] = mode;
            rc = Command(CommandCode.Request, 1, buf);

            if (rc != RetCode.MI_OK)
            {
                //printf("read card error code : " + rc.ToString("D3") + "\r\n");
            }
            else
            {
                int len = OutBuf[2];
                TagType = new Byte[len];
                for (int i = 0; i < len; i++)
                    TagType[i] = OutBuf[3 + i];
            }
            return rc;
        }

        private RetCode ReadCardNoA()
        {
            RetCode rc;

            rc = mifs_set_config();

            //   printf("rc = "+ +"\r\n");//test las

            if (rc != RetCode.MI_OK)
            {
                // printf("Error! RFID config Error Code : " + rc.ToString("X2") + "\r\n");
                return rc;
            }

            rc = mifs_request(1);    // 41 laslas

            if (rc != RetCode.MI_OK)
            {
                return rc;
            }

            rc = mifs_Anticoll2(0);   // 防止碰撞

            if (rc != RetCode.MI_OK)
            {
                return rc;
            }

            String NewCardNo = "";

            // 記錄下卡號
            for (int i = 0; i < 4; i++)
                NewCardNo += SNR[i].ToString("X2");

            if (CardNo.Length == 0 || NewCardNo.CompareTo(CardNo) != 0)
            {
                CardNo = "";
                // 記錄下卡號
                for (int i = 0; i < 4; i++)
                    CardNo += SNR[i].ToString("X2");

                CardType = "unknow";
                if (TagType[1] == 0)
                {
                    if (TagType[0] == 4)
                        CardType = "S50";
                    if (TagType[0] == 2)
                        CardType = "S70";
                }

                ShowLog(
                        "Found " + CardType + " Card!\r\n" +
                        "Card No :" + CardNo + "\r\n" +
                        "Size : 0x" + CardSize.ToString("X2") + "\r\n"
                       );
                TimeOut = false;
                // 開始進行計時
                tmr.Change(MaxOperationTime, Timeout.Infinite); 
            }
            return rc;
        }

        private RetCode Mifs_AuthKey(KeyType kt , int addr , Byte [] key)
        {
            Byte[] buf = new Byte[8];
            buf[0] = (Byte) kt;
            buf[1] = (Byte) (addr >> 2);
            for (int i=0;i<6;i++)
            {
                buf[i+2] = key[i];
            }

            return Command(CommandCode.AuthKey, 8, buf);  
        }

        private RetCode Mifs_Read(int addr)
        {
            Byte[] buf = new Byte[1];
            buf[0] = (Byte)addr;
            return Command(CommandCode.Read, 1, buf);   
        }

        private RetCode Mifs_Write(int addr)
        {
            Byte[] buf = new Byte[17];
            buf[0] = (Byte)addr;
            for (int i = 0; i < 16; i++)
            {
                if (Job == JobMode.AccessMode)
                    buf[i + 1] = AccessBlock[i];
                else
                    buf[i + 1] = WriteBuffer[i];
            }
            return Command(CommandCode.Write , 17 , buf);
        }

        private RetCode Mifs_IncValue(int addr)
        {
            Byte[] buf = new Byte[5];
            buf[0] = (Byte)addr;
            CovTool.UIntToByte(value, buf, 1);  
            return Command(CommandCode.Increment , 5, buf);
        }

        private RetCode Mifs_Transfer(int addr)
        {
            Byte[] buf = new Byte[1];
            buf[0] = (Byte)addr;
            return Command(CommandCode.Transfer, 1, buf);
        }

        private RetCode Mifs_DecValue(int addr)
        {
            Byte[] buf = new Byte[5];
            buf[0] = (Byte)addr;
            CovTool.UIntToByte(value, buf, 1);  
            return Command(CommandCode.Decrement, 5, buf);
        }

        private RetCode mifs_select()
        {
            RetCode rc = Command(CommandCode.Select,4,SNR);
            if (rc == RetCode.MI_OK)
            {
                int len = OutBuf[2];
                CardSize = OutBuf[3];
                ShowDebug("Card Size = " + CardSize + "\r\n"); 
            }
            return rc;
        }

        public enum IOCtrl
        {
             ReadCard = 1 ,
             Turn_OFF = 2
        }

        public enum RetCode  // Return Code
        {
            MI_OK = 0,
            MI_CHK_OK = 0,
            MI_CRC_ZERO = 0,
            MI_CRC_NOTZERO = 1,
            MI_NOTAGERR = 1,
            MI_CHK_FAILED = 1,
            MI_CRCERR = 2,
            MI_CHK_COMPERR = 2,
            MI_EMPTY = 3,
            MI_AUTHERR = 4,
            MI_PARITYERR = 5,
            MI_CODEERR = 6,

            MI_SERNRERR = 8,
            MI_KEYERR = 9,
            MI_NOTAUTHERR = 10,
            MI_BITCOUNTERR = 11,
            MI_BYTECOUNTERR = 12,
            MI_IDLE = 13,
            MI_TRANSERR = 14,
            MI_WRITEERR = 15,
            MI_INCRERR = 16,
            MI_DECRERR = 17,
            MI_READERR = 18,
            MI_OVFLERR = 19,
            MI_POLLING = 20,
            MI_FRAMINGERR = 21,
            MI_ACCESSERR = 22,
            MI_UNKNOWN_COMMAND = 23,
            MI_COLLERR = 24,
            MI_RESETERR = 25,
            MI_INITERR = 25,
            MI_INTERFACEERR = 26,
            MI_ACCESSTIMEOUT = 27,
            MI_NOBITWISEANTICOLL = 28,
            MI_QUIT = 30,
            MI_RECBUF_OVERFLOW = 50,
            MI_SENDBYTENR = 51,

            MI_SENDBUF_OVERFLOW = 53,
            MI_BAUDRATE_NOT_SUPPORTED = 54,
            MI_SAME_BAUDRATE_REQUIRED = 55,

            MI_WRONG_PARAMETER_VALUE = 60,

            MI_BREAK = 99,
            MI_NY_IMPLEMENTED = 100,
            MI_NO_MFRC = 101,
            MI_MFRC_NOTAUTH = 102,
            MI_WRONG_DES_MODE = 103,
            MI_HOST_AUTH_FAILED = 104,

            MI_WRONG_LOAD_MODE = 106,
            MI_WRONG_DESKEY = 107,
            MI_MKLOAD_FAILED = 108,
            MI_FIFOERR = 109,
            MI_WRONG_ADDR = 110,
            MI_DESKEYLOAD_FAILED = 111,

            MI_WRONG_SEL_CNT = 114,

            MI_WRONG_TEST_MODE = 117,
            MI_TEST_FAILED = 118,
            MI_TOC_ERROR = 119,
            MI_COMM_ABORT = 120,
            MI_INVALID_BASE = 121,
            MI_MFRC_RESET = 122,
            MI_WRONG_VALUE = 123,
            MI_VALERR = 124,
            MI_SAKERR = 0x46	//added by robbie
        }

        public enum CommandCode
        {
            Request = 0x41,
            Anticoll = 0x42,
            Anticoll2 = 0x71,
            CascAnticoll = 0x74,
            Select = 0x43,
            CascSelect = 0x75,
            Authentication = 0x44,
            Authentication2 = 0x72,
            AuthKey = 0x73,
            Halt = 0x45,
            Read = 0x46,
            Write = 0x47,
            ULWrite = 0x76,
            Increment = 0x48,
            Decrement = 0x49,
            Resore = 0x4A,
            Transfer = 0x4B,
            Value = 0x70,
            ValueDebit = 0x77,
            LoadKey = 0x4C,
            Reset = 0x4E,
            GetInfo = 0x4F,
            SetControlBit = 0x50,
            ClrControlBit = 0x51,
            Config = 0x52,
            WriteReg1 = 0x3D,
            ReadReg1 = 0x3E,
            Close = 0x3F,
            CheckWrite = 0x53,
            Buzzer = 0x60,
            ReadE2 = 0x61,
            WriteE2 = 0x62
        }
    }
}
